import java.util.*;
import java.time.*;
import java.time.temporal.*;
import java.time.chrono.*;
import java.time.format.*;

public class MyTime {
    public static void main(String... args) {
        // 共有コンテキストの準備(毎回作る必要はないので、アプリグローバルでも良い)
        ZoneId jst = ZoneId.of("JST", ZoneId.SHORT_IDS); // 日本（JST）のタイムゾーン
        //        ZoneId jst = ZoneId.systemDefault(); // 手元コードならこれでもOK
        ZoneId est = ZoneId.of("EST", ZoneId.SHORT_IDS); // 米国（EST）のタイムゾーン

        Locale jpLocale = new Locale("ja", "JP", "JP"); // 日本のロケール
        Chronology wareki = Chronology.ofLocale(jpLocale); // 和暦
        // Chronology wareki = Chronology.of("Japanese"); // 上記2行の代替例
        // ここまで共有コンテキストの準備

        // 入力数値からZonedDateTime(内部時刻)に変換
        // 想定例: 利用者は日本タイムゾーンで日時を入力
        ZonedDateTime myTime = ZonedDateTime.of(2014, 1, 31, 10/*hour*/, 0/*min*/, 0/*sec*/, 0/*nanosec*/, jst);

        // ZonedDateTime(内部時刻)から表示
        // 想定例: 利用者は日本タイムゾーンでの日時表示を期待
        System.out.format("%d, %d, %d\n", myTime.getYear(), myTime.getMonthValue(), myTime.getDayOfMonth());

        // ZonedDateTimeを使う日付計算
        // 単純な計算はplusMonthsやminusMonthsなどのメソッドを使う
        // 少し複雑な計算はTemporalAdjustersを使う
        // 例: 2014.1.31 =[1ヶ月後の最初の土曜日は?]=> 2014.3.1
        ZonedDateTime deadline = myTime.plusMonths(1).with(TemporalAdjusters.next(DayOfWeek.SATURDAY));
        // System.out.println(myTime.plusMonths(1).with(ChronoField.DAY_OF_WEEK, 6)); // 定数ベースの書き方(上記とは異なる意味です)

        // 時刻の比較
        // 例: 締切り前かの判定など
        if (myTime.compareTo(deadline) < 0) {
            System.out.println("before deadline");
        }

        // 日数計算
        // 例: 締切りまで残り何日か?
        // 注意: 締切り日の仕様を明確にするのはコード以前の話(たとえば、1月1日締切りの意味は、当日0時なのか夜11:59つまり事実上翌日の0時なのか)
        Duration duration = Duration.between(myTime, deadline);
        System.out.format("%d days until deadline\n", duration.toDays());

        // JST時刻からEST時刻に変換(Instantとしては同時刻)
        // 想定例: 日本のカレンダーで作った予定を米国のカレンダーで表示
        ZonedDateTime ustime = myTime.withZoneSameInstant(est);

        // ZonedDateTime(内部時刻)を和暦出力
        // 想定例: 利用者が和暦での日時表示を期待
        ChronoLocalDate heisei = wareki.date(myTime);
        // "Heisei"=>"平成"の文字列変換は自前実装が必要
        System.out.format("%s, %d, %d, %d\n", heisei.getEra(), heisei.get(ChronoField.YEAR_OF_ERA), heisei.get(ChronoField.MONTH_OF_YEAR), heisei.get(ChronoField.DAY_OF_MONTH));
        System.out.println(heisei.getEra().getDisplayName(TextStyle.FULL, jpLocale));

        // 和暦年号(H26)から西暦年号(2014年)へ変換
        // 想定例: 利用者が和暦で日時を入力
        String WAREKI_HEISEI = "Heisei"; // "Heisei"はハードコード
        JapaneseEra jpEra = JapaneseEra.valueOf(WAREKI_HEISEI);
        ChronoLocalDate heisei2 = wareki.date(jpEra, 26, 3, 15);
        System.out.println("A.D. " + heisei2.get(ChronoField.YEAR)); //=> 2014

        // タイムゾーンに依存しない日
        // 想定例: クリスマスはどのタイムゾーンのカレンダーで見てもずれない
        LocalDate xmasDay = LocalDate.of(2014, 12, 25);

        // 旧API(java.util.Date)との変換
        // java.util.Date => Date/Time API
        Date now = new Date();
        ZonedDateTime ztime = ZonedDateTime.ofInstant(now.toInstant(), jst);

        // Date/Time API => java.util.Date
        ZonedDateTime znow = ZonedDateTime.now();
        Date dateNow = Date.from(Instant.from(znow));
    }
}
