import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.EventHandler;
import javafx.geometry.Bounds;
import javafx.scene.control.SkinBase;
import javafx.scene.input.MouseEvent;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Arc;
import javafx.scene.shape.Circle;
import javafx.scene.shape.StrokeType;

public class KnobSkin extends SkinBase<Knob> {
    private Circle body;
    private Circle marker;
    private Arc highlight;
    private Arc shadow;
    private Circle minTick;
    private Circle maxTick;

    private final double totalAngle = 270d;
    private final double startAngle = (360d - totalAngle) / 2 + 90;
    private final double bodyRadius = 30d;
    private final Paint bodyColor = Color.LIGHTGRAY;
    private final double markerRadius = 3d;
    private final double markerSpacing = 4d;
    private final Paint markerColor = Color.GRAY;
    private final double tickRadius = 2d;
    private final double tickSpacing = 10d;
    private final Paint tickColor = markerColor;
    private final double borderWidth = 3d;
    private final double borderAngle = 45d;
    private final Paint highlightColor = Color.WHITE;
    private final Paint shadowColor = Color.GRAY;
    private final Paint borderColor = shadowColor;

    public KnobSkin(Knob knob) {
        super(knob);

        initChildren();
        initBehavior();
    }

    private void initBehavior() {
        getSkinnable().valueProperty().addListener(new ChangeListener<Number>() {
            @Override
            public void changed(ObservableValue<? extends Number> observable,
                                Number oldValue, Number newValue) {
                getSkinnable().requestLayout();
            }
        });

        getSkinnable().setOnMouseDragged(new EventHandler<MouseEvent>() {
            @Override
            public void handle(MouseEvent event) {
                Bounds contentBounds = getSkinnable().getLayoutBounds();
                double cx = contentBounds.getMinX() + contentBounds.getWidth() / 2;
                double cy = contentBounds.getMinY() + contentBounds.getHeight() / 2;
                
                double pointerX = event.getX() - cx;
                double pointerY = event.getY() - cy;
                double pointerAngle = Math.toDegrees(Math.atan2(pointerY, pointerX));
                
                double angle = pointerAngle - startAngle;
                if (pointerY < 0 || pointerX > 0) {
                    angle += 360;
                }
                
                double value = angle / totalAngle * (
                        getSkinnable().getMaxValue() - getSkinnable().getMinValue());
                value = Math.max(Math.min(value, getSkinnable().getMaxValue()),
                        getSkinnable().getMinValue());
                getSkinnable().setValue(value);
            }
        });
    }

    private void initChildren() {
        initBody();
        initMarker();
        initBorder();
        initTicks();

        getChildren().addAll(highlight, shadow, body, marker, minTick, maxTick);
    }

    private void initTicks() {
        minTick = new Circle(0, 0, tickRadius);
        minTick.setFill(tickColor);

        maxTick = new Circle(0, 0, tickRadius);
        maxTick.setFill(tickColor);
    }

    private void initBorder() {
        highlight = new Arc(0, 0, bodyRadius + borderWidth, bodyRadius + borderWidth,
                borderAngle, 180);
        highlight.setStrokeType(StrokeType.INSIDE);
        highlight.setStroke(borderColor);
        highlight.setFill(highlightColor);

        shadow = new Arc(0, 0, bodyRadius + borderWidth, bodyRadius + borderWidth,
                borderAngle + 180, 180);
        shadow.setStrokeType(StrokeType.INSIDE);
        shadow.setStroke(borderColor);
        shadow.setFill(shadowColor);
    }

    private void initMarker() {
        marker = new Circle(0, 0, markerRadius);
        marker.setFill(markerColor);
    }

    private void initBody() {
        body = new Circle(0, 0, bodyRadius);
        body.setFill(bodyColor);
    }

    @Override
    protected void layoutChildren(double contentX, double contentY,
                                  double contentWidth, double contentHeight) {
        double cx = contentX + contentWidth / 2;
        double cy = contentY + contentHeight / 2;

        layoutBody(cx, cy);
        layoutMarker(cx, cy);
        layoutBorder(cx, cy);
        layoutTicks(cx, cy);
    }

    private void layoutBorder(double cx, double cy) {
        double borderRadius = bodyRadius + borderWidth;
        double borderTheta = Math.toRadians(borderAngle);

        highlight.relocate(cx - borderRadius, cy - borderRadius);

        double shadowX = Math.cos(borderTheta) * borderRadius;
        double shadowY = Math.sin(borderTheta) * borderRadius;
        shadow.relocate(cx - shadowX, cy - shadowY);
    }

    private void layoutMarker(double cx, double cy) {
        double markerOrbitRadius = body.getRadius() - marker.getRadius() - markerSpacing;
        double markerAnglePerValue = totalAngle / (
                getSkinnable().getMaxValue() - getSkinnable().getMinValue());
        double markerAngle = startAngle + getSkinnable().getValue() * markerAnglePerValue;
        double markerTheta = Math.toRadians(markerAngle);
        double markerX = Math.cos(markerTheta) * markerOrbitRadius;
        double markerY = Math.sin(markerTheta) * markerOrbitRadius;

        marker.relocate(cx - marker.getRadius() + markerX, cy - marker.getRadius() + markerY);
    }

    private void layoutBody(double cx, double cy) {
        body.relocate(cx - body.getRadius(), cy - body.getRadius());
    }

    private void layoutTicks(double cx, double cy) {
        double tickOrbitRadius = bodyRadius + tickRadius + tickSpacing;

        double minTickTheta = Math.toRadians(startAngle);
        double minTickX = Math.cos(minTickTheta) * tickOrbitRadius;
        double minTickY = Math.sin(minTickTheta) * tickOrbitRadius;
        minTick.relocate(cx - tickRadius + minTickX, cy - tickRadius + minTickY);

        double maxTickTheta = Math.toRadians(startAngle + totalAngle);
        double maxTickX = Math.cos(maxTickTheta) * tickOrbitRadius;
        double maxTickY = Math.sin(maxTickTheta) * tickOrbitRadius;
        maxTick.relocate(cx - tickRadius + maxTickX, cy - tickRadius + maxTickY);
    }

    @Override
    protected double computeMinWidth(double height, double topInset,
                                     double rightInset, double bottomInset,
                                     double leftInset) {
        return computePrefWidth(height, topInset, rightInset, bottomInset, leftInset);
    }

    @Override
    protected double computeMaxWidth(double height, double topInset,
                                     double rightInset, double bottomInset,
                                     double leftInset) {
        return computePrefWidth(height, topInset, rightInset, bottomInset, leftInset);
    }

    @Override
    protected double computePrefWidth(double height, double topInset,
                                      double rightInset, double bottomInset,
                                      double leftInset) {
        return rightInset + tickRadius * 2 * 2 + tickSpacing * 2 + bodyRadius * 2 + leftInset;
    }

    @Override
    protected double computeMinHeight(double height, double topInset,
                                      double rightInset, double bottomInset,
                                      double leftInset) {
        return computePrefHeight(height, topInset, rightInset, bottomInset, leftInset);
    }

    @Override
    protected double computeMaxHeight(double height, double topInset,
                                      double rightInset, double bottomInset,
                                      double leftInset) {
        return computePrefHeight(height, topInset, rightInset, bottomInset, leftInset);
    }

    @Override
    protected double computePrefHeight(double width, double topInset,
                                       double rightInset, double bottomInset,
                                       double leftInset) {
        return topInset + tickRadius * 2 * 2 + tickSpacing * 2 + bodyRadius * 2 + bottomInset;
    }
}
